"""
Expose some internal functions through HTTP server, and rest API
"""
import logging

from aiohttp import web

log = logging.getLogger(__name__)


class ApiManager:
    """API"""

    def __init__(self, server):
        self.server = server

    async def start(self):
        """Start API"""
        log.debug("Build API")
        app = web.Application()
        app.router.add_get("/", self.handle_homepage)
        app.router.add_get("/load", self.handle_load)
        app.router.add_get("/dump", self.handle_dump)
        app.router.add_get("/flush", self.handle_flush)
        app.router.add_get("/limits", self.handle_get_limits)
        app.router.add_get("/periods", self.handle_get_periods)
        app.router.add_get("/reload_conf", self.handle_reload_configuration)
        app.router.add_get("/telegraf", self.handle_telegraf)
        runner = web.AppRunner(app)
        await runner.setup()

        site = web.TCPSite(
            runner,
            self.server.config.get("api", {}).get("api_host", "0.0.0.0"),  # nosec
            self.server.config.get("api", {}).get("api_port", 9002),
        )
        await site.start()
        log.info("API started")

    async def handle_homepage(self, request):
        """Handle request on API root"""

        # Handle legacy command request (specified with cmd parameter)
        command = request.query.get("cmd")
        if command:
            handler = {
                "load": self.handle_load,
                "dump": self.handle_dump,
                "flush": self.handle_flush,
                "get_limits": self.handle_get_limits,
                "get_periods": self.handle_get_periods,
                "reload_conf": self.handle_reload_configuration,
                "telegraf": self.handle_telegraf,
            }.get(command)
            if handler:
                return await handler(request, legacy=True)
            return web.Response(text="Bad request", status=400)

        return web.json_response({"success": False, "error": "Bad request"}, status=404)

    async def handle_load(self, request, legacy=False):
        """Load cache file content to RAM memory"""
        log.debug(
            "Handle load %s from %s", "legacy command" if legacy else "request", request.remote
        )
        self.server.data_manager.load()
        return web.Response(text="OK") if legacy else web.json_response({"success": True})

    async def handle_dump(self, request, legacy=False):
        """Dump cache content as JSON"""
        log.debug(
            "Handle dump %s from %s", "legacy command" if legacy else "request", request.remote
        )
        return web.json_response(self.server.data_manager.data)

    async def handle_flush(self, request, legacy=False):
        """Flush cached data to file"""
        log.debug(
            "Handle flush %s from %s", "legacy command" if legacy else "request", request.remote
        )
        self.server.data_manager.flush()
        return web.Response(text="OK") if legacy else web.json_response({"success": True})

    async def handle_reload_configuration(self, request, legacy=False):
        """Reload configuration file"""
        log.debug(
            "Handle reload configuration %s from %s",
            "legacy command" if legacy else "request",
            request.remote,
        )
        self.server.load_config()
        return web.Response(text="OK") if legacy else web.json_response({"success": True})

    async def handle_get_periods(self, request, legacy=False):
        """Get configured periods"""
        log.info(
            "Handle get periods %s from %s",
            "legacy command" if legacy else "request",
            request.remote,
        )
        return (
            web.Response(text=",".join(self.server.periods))
            if legacy
            else web.json_response(self.server.periods)
        )

    async def handle_get_limits(self, request, legacy=False):
        """Get user limits"""
        user = request.query.get("user", None)
        log.info(
            "Handle get user's limit %s from %s (user=%s)",
            "legacy command" if legacy else "request",
            request.remote,
            user,
        )

        if user is None:
            return (
                web.json_response(
                    {"success": False, "error": "User parameter is missing"}, status=400
                )
                if not legacy
                else web.Response(text="Unable to retrieve limit, missing user", status=500)
            )

        return web.json_response(self.__get_user_limits(user))

    async def handle_telegraf(self, request, legacy=False):
        """Get metric in Telegref JSON format"""
        log.info(
            "Handle telegraf %s from %s", "legacy command" if legacy else "request", request.remote
        )
        return web.json_response(
            {
                user: {
                    period: {
                        "count": (
                            float(usage.get(period, {}).get("count"))
                            if usage.get(period, {}).get("count")
                            else None
                        ),
                        "limit": float(limit) if float(limit) >= 0 else None,
                    }
                    for period, limit in self.__get_user_limits(user).items()
                }
                for user, usage in self.server.data_manager.data.items()
            }
        )

    def __get_user_limits(self, user):
        """Retrieve user limits from LDAP directory"""
        return {
            period: self.server.ldap_manager.get_limit_from_ldap(user, period)
            for period in self.server.periods
        }
