"""
Cache for data

Why ?
Because cache must be accessed by multiple users simultaneously

Guillaume Penaud
05/2015
"""

import json
import logging
import os
import os.path
import sys
import time
from datetime import datetime, timedelta

from postquota.ldap_manager import LdapManager

log = logging.getLogger(__name__)


class DataManager:
    """Manager for cached data"""

    def __init__(self, server):
        self.config = server.config
        self.periods = server.periods

        self.ldap_manager = LdapManager(server)

        self.last_save = time.time()
        self.save_every = int(self.config.get("data", {}).get("save_every", 60))

        # handle cache file
        cache_file = self.config.get("data", {}).get("file")

        if cache_file is None:
            log.warning("No path defined for cache file")
        elif not os.path.isfile(cache_file):
            os.makedirs(os.path.dirname(cache_file), exist_ok=True)
            with open(cache_file, "w", encoding="utf8") as file_desc:
                file_desc.write("{}")
            log.warning("Initialized an empty cache file: %s", cache_file)
        elif not os.access(cache_file, os.R_OK + os.W_OK):
            raise OSError(f"Insufficient permissions on cache file '{cache_file}'")

        self.cache_file = cache_file

        # set default data
        self.default_field_values = {"count": 0, "limit": 0}
        self.data = {}

        try:
            self.load()
        except Exception:  # pylint: disable=broad-exception-caught
            log.exception("Failed to load cache file (%s)", self.cache_file)

    def is_period_exceeded(self, user, period):
        """Determine if a period has been exceeded since last mail send"""
        log.debug("Check for period ttl exceeded for user %s at period %s", user, period)
        now_ts = datetime.now().timestamp()
        until_ts = self.data.get(user, {}).get(period, {}).get("until")

        if until_ts is None:
            return True

        if now_ts >= until_ts:
            return True

        return False

    def reset_count_for(self, user, period):
        """Reset cache count for one user on a period"""
        log.debug("Reset count of user %s for the period %s", user, period)

        until = {
            "hourly": lambda now: datetime(now.year, now.month, now.day, now.hour)
            + timedelta(hours=1),
            "daily": lambda now: datetime(now.year, now.month, now.day) + timedelta(days=1),
            "weekly": lambda now: datetime(now.year, now.month, now.day)
            + timedelta(days=7 - now.weekday()),
            "monthly": lambda now: (datetime(now.year, now.month, 1) + timedelta(days=32)).replace(
                day=1
            ),
            "yearly": lambda now: datetime(now.year + 1, 1, 1),
        }[period](datetime.now())

        data = self.data.setdefault(user, {}).setdefault(period, {})
        data["count"] = 0
        data["until"] = int(until.timestamp())

    def retrieve(self, user, period, field):
        """Get a field value of the main cache"""
        if self.is_period_exceeded(user, period):
            self.reset_count_for(user, period)

        field_value = (
            self.data.get(user, {}).get(period, {}).get(field, self.default_field_values[field])
        )

        return field_value

    def increment(self, user, value=1):
        """Increment count for a user"""
        log.debug("Increment cache for %s by %d", user, value)

        for period in self.periods:
            data = self.data.setdefault(user, {}).setdefault(period, {})
            count = data.setdefault("count", 0)
            data["count"] = count + value

    def load(self):
        """Populate cache from a file"""
        log.debug("Load from cache file")
        try:
            with open(self.cache_file, encoding="utf8") as file_desc:
                serialized_cache = file_desc.read()

            self.data = json.loads(serialized_cache)
        except ValueError:
            log.warning("Cache is empty")
            self.data = {}
        except OSError:
            log.error("Error while accessing cache file")
            self.data = {}

    def save(self):
        """Determine if this is time to flush cache or not"""
        if self.last_save and self.last_save > (time.time() - self.save_every):
            log.debug(
                "It's not yet time to flush cache to file (last saved: %s)",
                datetime.fromtimestamp(self.last_save),
            )
            return

        self.flush()

    def flush(self):
        """Flush cache content to file"""
        log.debug("Flush cache content to cache file")

        try:
            with open(self.cache_file, "w", encoding="utf8") as file_desc:
                file_desc.write(json.dumps(self.data))
            self.last_save = time.time()
        except OSError:
            log.error("Can not open cache file, please check")
            sys.exit(7)
