"""
Handle LDAP requests
"""

import logging
import time

import ldap

log = logging.getLogger(__name__)


class LdapManager:
    """Manage requests on LDAP directory"""

    def __init__(self, server):
        self.ldap_connection = None
        self.retry_count = 0
        self.server = server

        self.config = server.config
        self.periods = server.periods

        self.cache = {}

    def get_limit(self, user, period):
        """
        Get user period limit
        (from cache if available and non-expired, from LDAP otherwise)
        """
        cache_key = user + "-" + period
        if cache_key not in self.cache:
            self.get_limit_from_ldap(user, period)
        cache_entry = self.cache.get(cache_key)
        expiration = cache_entry.get("expiration")
        if expiration is None or expiration <= time.time():
            log.debug("Cache entry expired for user %s, regenerating...", user)
            self.get_limit_from_ldap(user, period)
        limit = self.get_limit_from_cache(user, period)
        return limit

    def get_limit_from_cache(self, user, period):
        """
        Tries to retrieve a value from cache entry - fallback to ldap request if an
        exception is triggered
        """
        log.debug("Retrieve limit from cache for user %s at period %s", user, period)
        limit = None

        cache_key = user + "-" + period
        cache_entry = self.cache.get(cache_key)

        if cache_entry is None:
            raise ValueError(f"No cache entry for cache_key: {user}-{period}")

        limit = cache_entry.get("limit")
        expiration = cache_entry.get("expiration")

        if limit is None:
            raise ValueError(f"Limit is unknown for user {user} and period {period}")

        if expiration is None or expiration <= time.time():
            raise ValueError(
                f"Cache entry expirated for user {user} and period {period}, regeneration..."
            )

        return limit

    def open_ldap_connection(self):
        """Check ldap connection, and initialize if needed"""
        log.debug("Open LDAP connection")
        uri = self.config.get("ldap", {}).get("uri", {})

        try:
            self.ldap_connection = ldap.initialize(uri)
            # pylint: disable=no-member
            self.ldap_connection.set_option(ldap.OPT_NETWORK_TIMEOUT, 10.0)

            bind_dn = self.config.get("ldap", {}).get("bind_dn", {})
            bind_pw = self.config.get("ldap", {}).get("bind_pw", {})

            if bind_dn is not None and bind_pw is not None:
                self.ldap_connection.simple_bind_s(bind_dn, bind_pw)

        except ldap.LDAPError as err:  # pylint: disable=no-member
            log.error(
                "Failed to establish connection to LDAP (attempt %d/5). Cause: %s",
                self.retry_count + 1,
                err,
            )
            if self.retry_count < 4:
                self.retry_count += 1
                time.sleep(5)
                self.open_ldap_connection()
            else:
                log.critical("Stopping after %s LDAP connection attempts", self.retry_count)
                self.retry_count = 0

    def get_limit_from_ldap(self, user, period_to_retrieve):
        """Refresh cache by ldap remote data"""
        log.debug("Retrieve limits from LDAP for user %s", user)
        try:
            self.open_ldap_connection()
            filters = self.config.get("ldap", {}).get("filter", "(uid=%s)") % user
            fields = [v_period["field"] for v_period in self.periods.values()]
            fields.append("uid")

            search_base = self.config.get("ldap", {}).get("search_base")

            result_id = self.ldap_connection.search(
                search_base, ldap.SCOPE_SUBTREE, filters, fields  # pylint: disable=no-member
            )
            result_type, result_data = self.ldap_connection.result(result_id, 2)

            if len(result_data) > 1:
                raise ValueError(
                    "LDAP search has returned ambiguous value, please change filter criteria"
                )

            if result_type == ldap.RES_SEARCH_ENTRY:  # pylint: disable=no-member
                ldap_data = result_data[0][1]

                for period, period_data in self.periods.items():
                    field = period_data.get("field", {})
                    limits = ldap_data.get(field)

                    if limits is None:
                        limit = "-1"
                    else:
                        limit = limits[0]

                    if isinstance(limit, bytes):
                        limit = limit.decode("utf-8")

                    self.update_limit_in_cache(user, period, limit)

            return self.get_limit_from_cache(user, period_to_retrieve)
        except ldap.LDAPError as err:  # pylint: disable=no-member
            log.error("Failed to fetch quota information from LDAP. Cause: %s", err)
            return -2

    def update_limit_in_cache(self, user, period, limit):
        """Refresh cache by retrieving ldap remote data"""
        log.debug("Update limit (%s) for user %s at period %s", limit, user, period)
        cache_key = user + "-" + period

        expiration = self.config.get("ldap", {}).get("cache_expiration_delay", 60)
        now = time.time()
        delay = now + int(expiration)

        # initialize hash if he does not already exists
        cache = self.cache.setdefault(cache_key, {})

        cache["limit"] = limit
        cache["expiration"] = delay
        cache["last_update"] = now

        return cache
