"""
Server that handle policy request from postfix
"""
import asyncio
import logging
import logging.handlers
import signal
import sys

import yaml

from postquota.api_manager import ApiManager
from postquota.data_manager import DataManager
from postquota.ldap_manager import LdapManager

log = logging.getLogger(__name__)


class Server:
    """Server that handle policy request from postfix"""

    reject_action = "REJECT"

    logger = None
    log_handler = None
    data_manager = None
    ldap_manager = None
    api_manager = None
    pol_server = None
    api_server = None

    def __init__(self, args=None):
        self.args = args or {}
        self.attrs = {}
        self.config_filepath = self.args.get(
            "configuration_file", "/etc/postquota/configuration.yaml"
        )
        self.config = {}
        self.periods = {}

    def load_config(self):
        """Load configuration object"""
        # Load configuration file
        try:
            with open(self.config_filepath, encoding="utf8") as file_desc:
                self.config = yaml.load(file_desc, Loader=yaml.SafeLoader)
        except OSError:
            sys.stderr.write("[ERROR] Impossible to load configuration file\n")
            sys.exit(1)

        # load logger here to be able to change configuration upon it
        self.init_logging()

        available_periods = ["hourly", "daily", "weekly", "monthly", "yearly"]
        periods = self.config.get("period")

        invalid_periods = [period for period in periods if period not in available_periods]
        if invalid_periods:
            log.error(
                "Invalid period(s) found in configuration: %s. Possible values: %s",
                ", ".join(invalid_periods),
                ", ".join(available_periods),
            )
            sys.exit(2)

        for label, data in sorted(
            periods.items(), key=lambda item: item[1]["nb_seconds"], reverse=True
        ):
            self.periods.setdefault(label, data)

        self.reject_action = self.config.get("global", {}).get("reject_action", self.reject_action)

    def init_logging(self):
        """Set-up custom logger"""
        if self.log_handler is not None:
            logging.getLogger().removeHandler(self.log_handler)

        log_level = self.config.get("global", {}).get("log_level", "INFO")
        logging.getLogger().setLevel(log_level)

        log_target = self.config.get("global", {}).get("log_target", "syslog")
        if log_target == "stderr":
            self.log_handler = logging.StreamHandler(sys.stderr)
        elif log_target == "stdout":
            self.log_handler = logging.StreamHandler(sys.stdout)
        else:
            mail_facility = logging.handlers.SysLogHandler.LOG_MAIL
            self.log_handler = logging.handlers.SysLogHandler(
                facility=mail_facility, address="/dev/log"
            )

        # to have a nice log format
        self.log_handler.setFormatter(
            logging.Formatter(
                " - ".join(
                    [
                        part
                        for part in [
                            "%(asctime)s" if log_target != "syslog" else None,
                            "%(module)s:%(lineno)d - %(funcName)s()"
                            if logging.getLogger().getEffectiveLevel() <= logging.DEBUG
                            else "%(name)s",
                            "%(levelname)s",
                            "%(message)s",
                        ]
                        if part
                    ]
                )
            )
        )
        logging.getLogger().addHandler(self.log_handler)

    async def handle_connection(self, reader, writer):
        """Handle client connections"""
        log.debug("Handle socket connection")
        while True:
            try:
                req = await reader.readuntil(separator=b"\n\n")
                try:
                    req = req.decode("utf8").strip()
                    log.debug("Received utf8 request:\n%s", req)
                except UnicodeError:
                    req = req.decode("iso-8859-1").strip()
                    log.warning("Received non-utf8 request:\n%s", req)

                attrs = self.parse_request(req)
                if "request" not in attrs:
                    log.error("Missing key 'request'.")

                res = f"action={self.smtpd_access_policy(attrs)}\n\n"
                writer.write(res.encode("utf-8") + b"\n\n")
                await writer.drain()
                log.debug("Sent response:\n%s", res)
            except asyncio.exceptions.IncompleteReadError as err:
                last_data_received = err.partial.decode("utf-8").strip()
                if last_data_received:
                    log.error(
                        "Remote side closed the connection. Last data received: '%s'",
                        last_data_received,
                    )
                else:
                    log.debug(
                        "Remote side closed the connection. Last data received: '%s'",
                        last_data_received,
                    )
                writer.close()
                break
            except Exception as err:  # pylint: disable=broad-exception-caught
                log.exception("Unexpected error while handling request", exc_info=True)
                writer.write(f"action=DUNNO -- Exception {str(err)}\n\n".encode())
                await writer.drain()
                break

    def add_signals_handler(self, loop, main_task):
        """Register signals handler"""
        for signum in [
            signal.SIGUSR1,
            signal.SIGUSR2,
            signal.SIGHUP,
            signal.SIGINT,
            signal.SIGTERM,
        ]:
            loop.add_signal_handler(
                signum, lambda signum=signum: self.handle_signal(signum, main_task)
            )

    def handle_signal(self, signum, main_task):
        """Handle user-defined signals"""
        if signum == signal.SIGUSR1:
            log.debug("Signal SIGUSR1 received - flushing cache")
            self.data_manager.flush()
        elif signum == signal.SIGUSR2:
            log.debug("Signal SIGUSR2 received - reloading configuration")
            self.load_config()
        elif signum == signal.SIGHUP:
            log.debug("Signal SIGHUP received - reloading cache")
            self.data_manager.load()
        elif signum in (signal.SIGINT, signal.SIGTERM):
            log.debug("Signal %s received - flush then stop daemon", signal.strsignal(signum))
            self.data_manager.flush()
            main_task.cancel()
        else:
            log.debug("Unknown signal %s received - pass", signal.strsignal(signum))

    def parse_request(self, data):
        """Parse incoming request"""
        attrs = {}
        for line in data.split("\n"):
            line = line.strip()
            if "=" not in line:
                log.warning("Ignoring invalid line: '%s'", line)
                continue
            key, val = (part.strip() for part in line.split("=", 1))
            if key in attrs:
                if not isinstance(attrs[key], list):
                    attrs[key] = [attrs[key]]
                attrs[key].append(val)
            else:
                attrs[key] = val
        return attrs

    def smtpd_access_policy(self, attrs):
        """Handle treatments relative to policy access"""
        log.debug("Handle mail policy")

        if "sasl_username" not in attrs:
            log.warning("Attr 'sasl_username' is missing. Ignoring.")
            return ""
        queue_id = attrs.get("queue_id", "?UNKNOWN-QUEUE-ID?")

        action = "DUNNO"
        messages = []
        increment = True

        user_with_fqdn = attrs.get("sasl_username")
        # we need to clean username here, to remove the domain part
        user = user_with_fqdn.split("@")[0]

        if user is None:
            log.info("No user, so the mail shall pass")
            return ""

        # Retrieve recipient_count
        # Note: The "recipient_count" attribute (Postfix 2.3 and later) is non-zero only in the
        # "DATA" and "END-OF-MESSAGE" stages.
        recipient_count = int(attrs.get("recipient_count", 1))

        for period in self.periods:
            count = self.data_manager.retrieve(user, period, "count")
            try:
                limit = self.ldap_manager.get_limit(user, period)
            except Exception:  # pylint: disable=broad-exception-caught
                log.warning(
                    "Unspecified limit for user=%s, period=%s. Considering 'unlimited'...",
                    user,
                    period,
                )
                continue

            if int(limit) == -2:
                # return "REJECT Service temporarily unavailable"
                continue

            # limit reached
            if int(limit) > -1 and int(count) - recipient_count > int(limit):
                increment = False
                action = f"{self.reject_action} {period} quota exhausted, please retry later"
                messages.append(f"[{period}: NO ({count + recipient_count}/{limit}])]")
                log.debug("Data for period: %s -- count:%s & limit:%s", period, count, limit)
                break

            messages.append(
                f"[{period}: YES "
                f"({count + recipient_count}/{'unlimited' if int(limit) == -1 else limit})]"
            )
            log.debug(
                "Data for period: %s -- count:%s & limit:%s",
                period,
                count,
                "unlimited" if int(limit) == -1 else limit,
            )

        # if an increment has been done previously, we need to try to save cache
        if increment:
            self.data_manager.increment(user, recipient_count)
            log.info("Mail %s accepted -- %s -- user=%s", queue_id, " - ".join(messages), user)
        else:
            log.info(
                "Mail %s rejected (%s) -- %s -- user=%s",
                queue_id,
                self.reject_action,
                " - ".join(messages),
                user,
            )

        self.data_manager.save()

        return f"{action} -- {queue_id} sasl_username='{user}' {' - '.join(messages)}"

    async def start(self):
        """Start server"""
        log.debug("Starting, please wait...")

        # load the managers
        self.data_manager = DataManager(self)
        self.ldap_manager = LdapManager(self)
        self.api_manager = ApiManager(self)

        self.pol_server = await asyncio.start_server(
            self.handle_connection,
            self.config.get("global", {}).get("policy_host", "127.0.0.1"),
            self.config.get("global", {}).get("policy_port", 9001),
        )
        await self.api_manager.start()

        async with self.pol_server:
            await self.pol_server.serve_forever()
